import os
import pandas as pd
from torchvision.datasets.folder import default_loader
from torchvision.datasets.utils import download_url
from torch.utils.data import Dataset

class CUB2011(Dataset):
	"""
	Caltech-UCSD Birds-200-2011
	This code is a version of https://github.com/TDeVries/cub2011_dataset.
	"""
	base_folder = 'CUB_200_2011/images'
	source_dir = 'CUB_200_2011'
	#url = 'http://www.vision.caltech.edu/visipedia-data/CUB-200-2011/CUB_200_2011.tgz'
	url = None # Currently, CUB_200_2011.tgz is located in Google drive.
	filename = 'CUB_200_2011.tgz'
	tgz_md5 = '97eceeb196236b17998738112f37df78'

	def __init__(self, root, train = True, transform = None, target_transform = None, download = False):
		"""
		Constructor
		"""
		self.root = root
		self.train = train
		self.transform = transform
		self.target_transform = target_transform
		self.loader = default_loader

		if download:
			self.download()

		if not self.check_integrity():
			msg = "The dataset is not found or has some errors. Please download it again."
			raise RuntimeError(msg)

	def load_configdata(self):
		"""
		Load the configuration data
		"""
		images = pd.read_csv(os.path.join(self.root, self.source_dir, 'images.txt'), sep = ' ', names = ['img_id', 'filepath'])
		image_class_labels = pd.read_csv(os.path.join(self.root, self.source_dir, 'image_class_labels.txt'), sep = ' ', names = ['img_id', 'target'])
		train_test_split = pd.read_csv(os.path.join(self.root, self.source_dir, 'train_test_split.txt'), sep = ' ', names = ['img_id', 'is_training_img'])
		data = images.merge(image_class_labels, on = 'img_id')
		self.data = data.merge(train_test_split, on = 'img_id')
		if self.train:
			self.data = self.data[self.data.is_training_img == 1]
		else:
			self.data = self.data[self.data.is_training_img == 0]

	def check_integrity(self):
		"""
		Verify the loaded data
		"""
		try:
			self.load_configdata()
		except Exception:
			return False

		for index, row in self.data.iterrows():
			filepath = os.path.join(self.root, self.source_dir, 'images', row.filepath)
			if not os.path.isfile(filepath):
				print(filepath)
				return False
			return True

	def download(self):
		"""
		Download files
		"""
		import tarfile
		if self.check_integrity():
			print('Files have already downloaded and verified')
			return
		if self.url is not None:
			download_url(self.url, self.root, self.filename, self.tgz_md5)
		with tarfile.open(os.path.join(self.root, self.filename), "r:gz") as tar:
			tar.extractall(path = self.root)

	def __len__(self):
		return len(self.data)

	def __getitem__(self, index, is_transformation = True):
		sample = self.data.iloc[index]
		path = os.path.join(self.root, self.source_dir, 'images', sample.filepath)
		target = sample.target - 1  # Targets start at 1 by default, so shift to 0
		img = self.loader(path)

		if (self.transform is not None) and (is_transformation):
			img = self.transform(img)

		return img, target

if __name__ == '__main__':
	dataset = CUB2011(root = '../datasets/', train = True, transform = None, target_transform = None, download = True)
	print("Description DegradedCIFAR10 dataset (Not Tensor):")
	print(dataset)
	out_data = dataset.__getitem__(index=1)
	print("Save images into test_imgs from:")
	print(out_data)

	(img, target) = out_data
	img.save('../test_imgs/CUB2011_img.png')
