import torch
import torchvision
from torchvision import transforms
from torchvision.transforms import functional as tvtf
from PIL import Image

from datahandler import degtransforms

class DegradedCIFAR100(torchvision.datasets.CIFAR100):
	"""
	Degraded CIFAR100 class
	Both 'transform' and 'target_transform' do not include 'ToTensor' class.
	This class will apply an image degradation after calling 'transform' or 'target_transform'.
	If 'transform' or 'target_transform' include 'ToTensor' class, 'ToTensor' will be done in the end.
	It means that an image degradation is prior to 'ToTensor'.
	"""
	def __init__(self, root, train = True, transform = None, target_transform = None, download = False, deg_type = "jpeg", deg_range = None, deg_list = None, is_to_tensor = False, is_target_to_tensor = False, is_normalized = False, is_hflip = False):
		"""
		Constructor
		Regarding parameters not described here, please refer the document of CIFAR10 class.
		Args:
			deg_type (string) : degradation type
			deg_range (list) : the range of degradation levels [deg_start, deg_end]
			deg_list (list) : the list of degradation levels [deg_lev1, deg_lev2, deg_lev3,...]
			is_to_tensor : True if inputs are converted to Tensor
			is_target_to_tensor : True if targets are converted to Tensor
			is_normalized : Apply the normalization seen in https://github.com/dyhan0920/PyramidNet-PyTorch/blob/master/train.py
			is_hflip : True apply Holizontalflip
		"""
		super().__init__(root, train, transform, target_transform, download)
		self.deg_type = deg_type
		self.deg_range = deg_range
		self.deg_list = deg_list
		self.deg_transform = transforms.Compose([degtransforms.DegradationApplyWithLevel(self.deg_type, self.deg_range, self.deg_list)])
		self.is_to_tensor = is_to_tensor
		self.is_normalized = is_normalized
		self.is_hflip = is_hflip
		if is_to_tensor:
			if self.is_normalized:
				normalize = transforms.Normalize(mean = [129.3/255.0, 124.1/255.0, 112.4/255.0], std = [68.2/255.0, 65.4/255.0, 70.4/255.0])
				self.deg_to_tensor = transforms.Compose([transforms.ToTensor(), normalize])
			else:
				self.deg_to_tensor = transforms.Compose([transforms.ToTensor()])
		else:
			self.deg_to_tensor = None
		self.is_target_to_tensor = is_target_to_tensor

		transform_str = transform.__repr__()
		target_transform_str = target_transform.__repr__()

		if ('ToTensor' in transform_str) or ('ToTensor' in target_transform_str):
			msg = "Both transform and target_transform should not iclude ToTensor. If you iclude tensor, please use is_to_tensor and is_target_to_tensor."
			raise TypeError(msg)

	def __getitem__(self, index):
		"""
		degradation & tensor are applied.
		"""
		clean_img, target = super().__getitem__(index)
		deg_img, deg_lev = self.deg_transform(clean_img)
		if self.train: # self.transform does not have the RandomCrop in the training process
			#result = Image.new(clean_img.mode, (32+4+4, 32+4+4), (0, 0, 0))
			#result.paste(clean_img, (4, 4))
			#clean_img = result
			#result_deg = Image.new(deg_img.mode, (32+4+4, 32+4+4), (0, 0, 0))
			#result_deg.paste(deg_img, (4, 4))
			#deg_img = result_deg
			clean_img = tvtf.pad(clean_img, 4, 0, "constant")
			deg_img = tvtf.pad(deg_img, 4, 0, "constant")
			i, j, h, w = transforms.RandomCrop.get_params(clean_img, output_size=(32, 32))
			clean_img = tvtf.crop(clean_img, i, j, h, w)
			deg_img = tvtf.crop(deg_img, i, j, h, w)
			
			if self.is_hflip:
				if torch.rand(1) < 0.5: # probability = 0.5
					clean_img = tvtf.hflip(clean_img)
					deg_img = tvtf.hflip(deg_img)

		if self.is_to_tensor:
			tensor_clean_img = self.deg_to_tensor(clean_img)
			tensor_deg_img = self.deg_to_tensor(deg_img)
			imgs = (tensor_clean_img, tensor_deg_img)
		else:
			imgs = (clean_img, deg_img)

		if self.is_target_to_tensor:
			deg_lev = degtransforms.mormalize_level(self.deg_type, deg_lev)
			tensor_target = torch.tensor(target)
			#tensor_deg_lev = torch.tensor(deg_lev)
			tensor_deg_lev = deg_lev
			targets = (tensor_target, tensor_deg_lev)
		else:
			targets = (target, deg_lev)

		return imgs, targets

	def getCHW(self):
		"""
		Return (Channel, Height, Width)
		"""
		return (3,32,32)
