import torch
import torchvision
from torchvision import transforms
from torchvision.transforms import functional as tvtf

from datahandler import degtransforms
from datahandler import cub2011

class DegradedCUB2011(cub2011.CUB2011):
	"""
	Degraded Caltech-UCSD Birds-200-2011
	Both 'transform' and 'target_transform' do not include 'ToTensor' class.
	This class will apply an image degradation after calling 'transform' or 'target_transform'.
	If 'transform' or 'target_transform' include 'ToTensor' class, 'ToTensor' will be done in the end.
	It means that an image degradation is prior to 'ToTensor'.
	This code is a version of https://github.com/TDeVries/cub2011_dataset.
	"""
	def __init__(self, root, train = True, transform = None, target_transform = None, download = False, deg_type = "jpeg", deg_range = None, deg_list = None, is_to_tensor = False, is_target_to_tensor = False, is_ten_crop = False, final_crop_size = 224, is_normalized = False, is_resized_before_deg = True, cgn_traindata_gen = False):
		"""
		Constructor
		Regarding parameters not described here, please refer the document of CIFAR10 class.
		Args:
			deg_type (string) : degradation type
			deg_range (list) : the range of degradation levels [deg_start, deg_end]
			deg_list (list) : the list of degradation levels [deg_lev1, deg_lev2, deg_lev3,...]
			is_to_tensor (bool) : True if inputs are converted to Tensor
			is_target_to_tensor (bool) : True if targets are converted to Tensor (Selecting True always make error occur.)
			is_ten_crop (bool) : True if the ten cropping transform is applied
			final_crop_size (int) : the final size of cropping
			is_normalized (bool) : True if images are normalized by the torchvision manner only. It works only when is_to_tensor is True.
			is_resized_before_deg (bool) : True if degradation is applied after resizing orginal images.
			cgn_traindata_gen (bool) : True, Special data generation for CGN experiments
		"""
		super().__init__(root, train, transform, target_transform, download)
		self.deg_type = deg_type
		self.deg_range = deg_range
		self.deg_list = deg_list
		self.deg_transform = transforms.Compose([degtransforms.DegradationApplyWithLevel(self.deg_type, self.deg_range, self.deg_list)])
		self.is_to_tensor = is_to_tensor
		if is_to_tensor:
			if is_normalized:
				self.deg_to_tensor = transforms.Compose([transforms.ToTensor(), transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))])
			else:
				self.deg_to_tensor = transforms.Compose([transforms.ToTensor()])
		else:
			self.deg_to_tensor = None
		self.is_target_to_tensor = is_target_to_tensor
		self.is_ten_crop = is_ten_crop
		self.final_crop_size = final_crop_size
		self.is_normalized = is_normalized
		self.is_resized_before_deg = is_resized_before_deg
		self.cgn_traindata_gen = cgn_traindata_gen
		if self.cgn_traindata_gen:
			self.is_resized_before_deg = False

		transform_str = transform.__repr__()
		target_transform_str = target_transform.__repr__()

		if ('ToTensor' in transform_str) or ('ToTensor' in target_transform_str):
			msg = "Both transform and target_transform should not iclude ToTensor. If you iclude tensor, please use is_to_tensor and is_target_to_tensor."
			raise TypeError(msg)
		if not(self.is_resized_before_deg):
			if self.train:
				if ('Random' in transform_str) or ('TenCrop' in transform_str):
					msg = "Do not iclude random data augmentationa in transform."
					raise TypeError(msg)
			else:
				if ('CenterCrop' in transform_str) or ('TenCrop' in transform_str):
					msg = "Do not iclude the CenterCrop in transform."
					raise TypeError(msg)

	def __getitem__(self, index):
		"""
		degradation & tensor are applied.
		"""
		clean_img_orgsize, target = super().__getitem__(index, is_transformation = self.is_resized_before_deg)

		if self.is_resized_before_deg: # have already been resized.
			clean_img = clean_img_orgsize
			deg_img, deg_lev = self.deg_transform(clean_img) # Degradation
		else:
			if self.train: # self.transform does not have the RandomHorizotanlFlip in the training process
				clean_img_orgsize = transforms.RandomHorizontalFlip()(clean_img_orgsize)
			# Degradation
			deg_img_orgsize, deg_lev = self.deg_transform(clean_img_orgsize)
			# Cropping 256
			clean_img = self.transform(clean_img_orgsize)
			deg_img = self.transform(deg_img_orgsize)

		if self.is_ten_crop: # degradatin level transformation is not applied
			if self.is_to_tensor:
				if self.is_normalized:
					tentrans = transforms.Compose([transforms.TenCrop(self.final_crop_size), transforms.Lambda(lambda crops: torch.stack([transforms.ToTensor()(crop) for crop in crops])), transforms.Lambda(lambda tensors:torch.stack([transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])(t) for t in tensors]))])
				else:
					tentrans = transforms.Compose([transforms.TenCrop(self.final_crop_size), transforms.Lambda(lambda crops: torch.stack([transforms.ToTensor()(crop) for crop in crops]))])
				tensor_clean_img = tentrans(clean_img)
				tensor_deg_img = tentrans(deg_img)
				imgs = (tensor_clean_img, tensor_deg_img)
			else:
				tentrans = transforms.Compose([transforms.TenCrop(self.final_crop_size)])
				clean_img = tentrans(clean_img)
				deg_img = tentrans(deg_img)
				imgs = (clean_img, deg_img)
		else:
			if not(self.is_resized_before_deg):
				if self.train: # self.transform does not have the RandomCrop in the training process
					i, j, h, w = transforms.RandomCrop.get_params(clean_img, output_size=(self.final_crop_size, self.final_crop_size))
					inp_clean_img = tvtf.crop(clean_img, i, j, h, w)
					deg_img = tvtf.crop(deg_img, i, j, h, w)
				else:
					inp_clean_img = transforms.CenterCrop(size = self.final_crop_size)(clean_img)
					deg_img = transforms.CenterCrop(size = self.final_crop_size)(deg_img)

				if self.cgn_traindata_gen:
					deg_lev = degtransforms.calc_deglev_rescaledimg(self.deg_type, clean_img_orgsize, clean_img, inp_clean_img, deg_img, deg_lev)

			if self.is_to_tensor:
				tensor_clean_img = self.deg_to_tensor(inp_clean_img)
				tensor_deg_img = self.deg_to_tensor(deg_img)
				imgs = (tensor_clean_img, tensor_deg_img)
			else:
				imgs = (inp_clean_img, deg_img)

		if self.is_target_to_tensor:
			deg_lev = degtransforms.mormalize_level(self.deg_type, deg_lev)
			tensor_target = torch.tensor(target)
			tensor_deg_lev = deg_lev
			targets = (tensor_target, tensor_deg_lev)
		else:
			targets = (target, deg_lev)

		return imgs, targets

	def getCHW(self):
		"""
		Return (Channel, Height, Width)
		"""
		return (3,224,224)
