import torch
import torchvision
from torchvision import transforms
from datahandler import degradedcifar10, degradedcifar100, degradedcub2011, degradedcamvid
import numpy as np

def get_type_range(degtype):
	"""
	Get degradation type and range
	Args:
		degtype (string) : jpeg, noise, blur, saltpepper
	"""
	if degtype == "jpeg":
		deg_type = "jpeg"
		deg_range = [1, 101]
	elif degtype == "noise":
		deg_type = "noise"
		deg_range = [0, 50]
	elif degtype == "blur":
		deg_type = "blur"
		deg_range = [0, 50]
	else:
		deg_type = "saltpepper"
		deg_range = [0, 25]
	return deg_type, deg_range

def get_minmax_normalizedlevel(deg_type):
	"""
	Min and Max of normalized degradation levels
	Args:
		deg_type (string) : degradation type
	Returns:
		normalized degradation level (tuple)
	"""
	if deg_type == 'jpeg':
		ret_adj = 100.0
		max_l = 101.0
		min_l = 1.0
	elif deg_type == 'noise':
		ret_adj = 255.0
		max_l = 50.0
		min_l = 0.0
	elif deg_type == 'blur':
		ret_adj = 100.0
		max_l = 50.0
		min_l = 0.0
	elif deg_type == 'saltpepper':
		ret_adj = 100.0
		max_l = 5.0
		min_l = 0.0
	else:
		ret_adj = 1.0
		max_l = 1.0
		min_l = 0.0

	return min_l/ret_adj, max_l/ret_adj

def fix_seed_noise_sl(is_fixed):
	"""
	Fix the seed of Gaussian and Binomial distributions
	This is only used for evalution purpose.
	If you fix the seed, please do not forget to unfix the seed.
	Args:
		is_fixed (bool) : True if the seed is fixed
	"""
	if is_fixed:
		np.random.seed(seed=301)
	else:
		np.random.seed(seed=None)

def get_datasetclass(data_name = "CIFAR10", train = True, transform = None, deg_type = "jpeg", deg_range = [1,101], deg_list = None, download = True, is_resized_before_deg = True, cgn_traindata_gen = False):
	"""
	Get a dataset object
	Args:
		data_name (string) : dataset name
		train (bool) : True if training data, False if test data
		transform (transforms) : data augmentation
		deg_type (string) : jpeg, noise, blur, saltpepper
		deg_range (list) : degradation levels (range type)
		deg_list (list) : degradation levels (list type)
	"""
	if data_name == "CIFAR10":
		datasetclass = degradedcifar10.DegradedCIFAR10( root = './datasets/', train = train, download = download, transform = transform, deg_type = deg_type, deg_range = deg_range, deg_list = deg_list, is_to_tensor = True, is_target_to_tensor = True, is_normalized = True)
	elif data_name == "CIFAR100":
		datasetclass = degradedcifar100.DegradedCIFAR100( root = './datasets/', train = train, download = download, transform = transform, deg_type = deg_type, deg_range = deg_range, deg_list = deg_list, is_to_tensor = True, is_target_to_tensor = True, is_normalized = True, is_hflip = False)
	elif data_name == "CUB2011":
		datasetclass = degradedcub2011.DegradedCUB2011( root = './datasets/', train = train, download = download, transform = transform, deg_type = deg_type, deg_range = deg_range, deg_list = deg_list, is_to_tensor = True, is_target_to_tensor = True, is_normalized = True, is_resized_before_deg = is_resized_before_deg, cgn_traindata_gen = cgn_traindata_gen)
	else:
		datasetclass = None
	return datasetclass

def get_type_list(degtype):
	"""
	Get degradation type and range
	Args:
		degtype (string) : jpeg, noise, blur, saltpepper
	"""
	if degtype == "jpeg":
		deg_type = "jpeg"
		deg_list = [10, 30, 50, 70, 90]
	elif degtype == "noise":
		deg_type = "noise"
		deg_list = [np.sqrt(0.05)*255, np.sqrt(0.1)*255, np.sqrt(0.15)*255, np.sqrt(0.2)*255, np.sqrt(0.25)*255]
	elif degtype == "blur":
		deg_type = "blur"
		deg_list = [10., 20., 30., 40., 50.]
	else:
		deg_type = "saltpepper"
		deg_list = [5., 10., 15., 20., 25.]

	return deg_type, deg_list
