"""
Common functions for Degradation Removal Adapter
"""
import os
import torch
import torch.nn as nn
import csv
import numpy as np
from datetime import datetime

def save_model(root_dir = None, model_name = 'dramodel', epoch = 0, target_dict = None):
	"""
	Save a model checkpoint as *.pth
	Args:
		root_dir (string) : path name
		model_name (string) : The file name is '{model_name}_{epoch}.pth.'
		epoch (int) : epoch number
		target_dict (dictionary) : model checkpoint dictionary
	"""
	file_name = '{}_{}.pth'.format(model_name, epoch)
	path = os.path.join(root_dir, file_name)
	os.makedirs(root_dir, exist_ok=True)
	torch.save(target_dict, path)

def load_model(root_dir = None, model_name = 'dramodel', epoch = 0):
	"""
	Load a model checkpoint
	Args:
		root_dir (string) : path name
		model_name (string) : The file name is '{model_name}_{epoch}.pth.'
		epoch (int) : epoch number
	"""
	file_name = '{}_{}.pth'.format(model_name, epoch)
	path = os.path.join(root_dir, file_name)
	if os.path.exists(path):
		checkpoint = torch.load(path)
	else:
		checkpoint = None
	return checkpoint

def save_csv(root_dir = None, file_name = None, target_list = None, header_list = None):
	"""
	Save the training progress
	Args:
		root_dir (string) : path name
		file_name (string) : The file name is '{file_name}.csv.'
		target_list : A list should be output.
	"""
	path = os.path.join(root_dir, file_name)
	if not(os.path.exists(path)):
		os.makedirs(root_dir, exist_ok=True)
		with open(path, 'a') as csvfile:
			writer = csv.writer(csvfile, lineterminator='\n')
			if header_list is not None:
				writer.writerow(header_list)
			writer.writerow(target_list)
	else:
		with open(path, 'a') as csvfile:
			writer = csv.writer(csvfile, lineterminator='\n')
			writer.writerow(target_list)

def get_logmodel_dir(deg_type = "jpeg", model_name = None, deg_flag = None, deg_lev_est = False, feature_loss = None, simul = False, exp_string = "CLEAN", rootname = 'models'):
	"""
	Get the directory of saved models
	Args:
		deg_type (string) : jpeg, noise, blur, saltpepper
		model_name (string) : VGG16, Resnet, ...
		deg_flag (int or None) : None clean only, 0 simple distillation, 1 degradation removal adapter.
		deg_lev_est (bool) : True in multi-task learning of degradation levels.
		feature_loss (string) : The name of a feature loss function
		simul (bool) : True if training simultenously
		exp_string (string) : "CLEAN" or "DEG"
		rootname (string) : root directory's name
	"""
	if simul:
		consistency = "simul"
	elif deg_flag == 0:
		if deg_lev_est:
			consistency = "deglev_distil"
		else:
			consistency = "simple_distil"
	elif deg_flag == 1:
		if deg_lev_est:
			consistency = "adapter_deglev"
		else:
			consistency = "adapter_only"
	elif deg_flag == None:
		if exp_string == "CLEAN":
			consistency = "clean"
		elif exp_string == "DEG":
			consistency = "degrade"
	else:
		msg = 'Unsupported degradation flag.'
		raise TypeError(msg)

	if feature_loss is None:
		dir = os.path.join(rootname, deg_type, model_name, consistency)
	else:
		dir = os.path.join(rootname, deg_type, model_name, consistency, feature_loss)

	return dir

def num_correctness(outputs, labels):
	"""
	Return the number of correct predictions
	Args:
	outputs (tensor) : predictions
	labels (tensor) : correct labels
	"""
	return (outputs.max(1)[1] == labels).sum().item()

def timestamp():
	"""
	Timestamp
	"""
	return datetime.now().strftime("%Y/%m/%d %H:%M:%S")

def length_and_degree(target_tensor1, target_tensor2):
	"""
	Calculate the lengths and degree of two tensors
	Args:
		target_tensor1 (tensor) : tensor1
		target_tensor2 (tensor) : tensor2
	Return:
		L2 norm of tensor1 (tensor)
		L2 norm of tensor2 (tensor)
		degreee between tensor1 and tensor2 (tensor)
		length difference between tensor1 and tensor2 (tensor)
	"""
	cos = nn.CosineSimilarity()
	deg_12 = cos(target_tensor1, target_tensor2)
	deg_12 = torch.acos(deg_12)
	deg_12 = torch.mul(deg_12, 180.0/np.pi)
	norm_1 = torch.norm(target_tensor1)
	norm_2 = torch.norm(target_tensor2)
	diff_12 = torch.sub(norm_1, norm_2)
	diff_12 = torch.abs(diff_12)
	return norm_1, norm_2, deg_12, diff_12

def get_eval_dir_filename(deg_type = "jpeg", model_name = None, deg_flag = None, deg_lev_est = False, feature_loss = None, simul = False, is_feat_only = True, epoch = None, exp_string = "CLEAN", rootname = 'evaluations'):
	"""
	Get the directory and filename of evaluations
	Args:
		deg_type (string) : jpeg, noise, blur, saltpepper
		model_name (string) : VGG16, Resnet, ...
		deg_flag (int or None) : None clean only, 0 simple distillation, 1 degradation removal adapter.
		deg_lev_est (bool) : True in multi-task learning of degradation levels.
		feature_loss (string) : The name of a feature loss function
		simul (bool) : True if training simultenously
		is_feat_only (bool) : True if train only features, False if features and task are trained.
		epoch : epoch number
		exp_string (string) : "CLEAN" or "DEG"
		rootname (string) : root directory's name
	Return:
		dir, file_name
	"""
	if simul:
		consistency = "simul"
	elif deg_flag == 0:
		if deg_lev_est:
			consistency = "deglev_distil"
		else:
			consistency = "simple_distil"
	elif deg_flag == 1:
		if deg_lev_est:
			consistency = "adapter_deglev"
		else:
			consistency = "adapter_only"
	elif deg_flag == None:
		if exp_string == "CLEAN":
			consistency = "clean"
		elif exp_string == "DEG":
			consistency = "degrade"
	else:
		msg = 'Unsupported degradation flag.'
		raise TypeError(msg)

	file_name = "res_" + consistency
	if feature_loss is not None:
		file_name += ("_" + feature_loss)
		if is_feat_only:
			file_name += "_feat"
		else:
			file_name += "_all"

	if epoch is not None:
		file_name += "_"
		file_name += str(epoch)
	file_name += ".csv"

	dir = os.path.join(rootname, deg_type, model_name)

	return dir, file_name

if __name__ == '__main__':
	print('----- TEST START -----')
	d = ["aaa", 0.1, 123, "abc"]
	save_csv(root_dir = 'dra_common', file_name = 'dramodel.csv', target_list = d)
	d = ["2aaa", 0.1, 123, "abc"]
	save_csv(root_dir = 'dra_common', file_name = 'dramodel.csv', target_list = d)
	outdir = get_logmodel_dir(deg_type = "jpeg", model_name = "VGG", deg_flag = None, deg_lev_est = False)
	assert outdir == 'models/jpeg/VGG/clean', '{}!={}'.format(outdir, 'models/jpeg/VGG/clean')
	outdir = get_logmodel_dir(deg_type = "jpeg", model_name = "VGG", deg_flag = 0, deg_lev_est = False)
	assert outdir == 'models/jpeg/VGG/simple_distil', '{}!={}'.format(outdir, 'models/jpeg/VGG/simple_distil')
	outdir = get_logmodel_dir(deg_type = "jpeg", model_name = "VGG", deg_flag = 1, deg_lev_est = False)
	assert outdir == 'models/jpeg/VGG/adapter_only', '{}!={}'.format(outdir, 'models/jpeg/VGG/adapter_only')
	outdir = get_logmodel_dir(deg_type = "jpeg", model_name = "VGG", deg_flag = 1, deg_lev_est = True)
	assert outdir == 'models/jpeg/VGG/adapter_deglev', '{}!={}'.format(outdir, 'models/jpeg/VGG/adapter_deglev')
	outdir = get_logmodel_dir(deg_type = "jpeg", model_name = "VGG", deg_flag = 2, deg_lev_est = True)
	assert outdir == 'models/jpeg/VGG/simul', '{}!={}'.format(outdir, 'models/jpeg/VGG/simul')
	try:
		outdir = get_logmodel_dir(deg_type = "jpeg", model_name = "VGG", deg_flag = 10, deg_lev_est = True)
	except TypeError:
		print('Done without any errors.')
