"""
Evaluate Degradation Removal Adapter
"""
import sys
import torch
import torch.nn as nn
import torchvision
from torchvision import transforms
from tqdm import tqdm
import json
import numpy as np
import copy
import torch.nn.functional as F

from modelforcifar import dra_nwbuilder
from datahandler import degradedimagedata
from dra_common import save_model, load_model, save_csv, get_logmodel_dir, num_correctness, timestamp, length_and_degree, get_eval_dir_filename

# Sys args
last_epoch = sys.argv[1] # Last epoch, should be "0" in the case of initial training.
exp_string = sys.argv[2] # Select a key string from "CLEAN", "MSE" and "COS"
exp_no = int(sys.argv[3]) # 0: CLEAN, 1-4: MSE, 1-8: COS

# Meta settings
batch_size = 100
num_workers = 2
simul = False # False if sequential or Ture if simultenous
headerlist=["Deg_level","Est_a","Est_b","Est_level","Est_a_std","Est_b_std","Est_level_std"]

# Configuration for experiments
model_flag = 0
if model_flag == 0:
	open_config = open('experiments/vgg16cifar10_config.json','r')
elif model_flag == 1:
	open_config = open('experiments/resnet50cifar10_config.json','r')
elif model_flag == 2:
	open_config = open('experiments/shakepyramidcifar10_config.json','r')

try:
	model_config = json.load(open_config)
finally:
	open_config.close()

d = model_config[exp_string][exp_no]
deg_type, deg_range = degradedimagedata.get_type_range(d["deg_type"])
deg_flag = d["deg_flag"]

dir, filename = get_eval_dir_filename(deg_type = d["deg_type"], model_name = d["model_name"], deg_flag = deg_flag, deg_lev_est = d["deg_lev_est"], feature_loss = d["feature_loss"], simul = simul, is_feat_only = d["is_feat_only"], epoch = last_epoch, exp_string = exp_string)
filename = 'ABD_' + filename

# load a model
device = 'cuda' if torch.cuda.is_available() else 'cpu'
root_dir = get_logmodel_dir(deg_type = d["deg_type"], model_name = d["model_name"], deg_flag = deg_flag, deg_lev_est = d["deg_lev_est"], feature_loss = d["feature_loss"], simul = simul, exp_string = exp_string)
last_checkpoint = load_model(root_dir = root_dir, epoch = last_epoch)
dra_nw = dra_nwbuilder.buildNetwork(model_name = d["model_name"], num_of_classes = d["num_classes"], num_of_features = d["num_features"], init_weights = True, deg_flag = d["deg_flag"], is_deglev_estimated = d["deg_lev_est"])
dra_nw.load_state_dict(last_checkpoint["model_state_dict"])
net = dra_nw.to(device)
net.eval()

# Loss function
cross_entoropy_loss_clean = nn.CrossEntropyLoss()
cross_entoropy_loss_deg = nn.CrossEntropyLoss()
if d["feature_loss"] == "MSE":
	consistency_loss = nn.MSELoss()
elif d["feature_loss"] == "COS":
	consistency_loss = nn.CosineEmbeddingLoss()
elif d["feature_loss"] == "KLD":
	consistency_loss = nn.KLDivLoss(reduction='batchmean')
elif d["feature_loss"] == "DIS":
	consistency_loss = nn.KLDivLoss(reduction='batchmean')
elif d["feature_loss"] is not None:
	msg = 'The type of consitency loss is not supported.'
	raise LookupError(msg)

if d["deg_lev_est"]:
	deg_levels_loss = nn.MSELoss()

# Fix the seeds of Guassian and binomial distributions for degradation
degradedimagedata.fix_seed_noise_sl(is_fixed = True)

print("----- Evaluation START -----")
print(*headerlist)

for lev in range(deg_range[0],deg_range[1]+1):
	# Data generation
	test_dataset = degradedimagedata.get_datasetclass(data_name = d["data_name"], train = False, transform = None, deg_type = deg_type, deg_range = [lev, lev], deg_list = None, download = False)
	test_loader = torch.utils.data.DataLoader( test_dataset, batch_size = 1, shuffle = False, num_workers = num_workers)

	# Initialize variabels
	val_est_a = 0
	val_est_b = 0
	val_est_level = 0
	val_est_a_std = 0
	val_est_b_std = 0
	val_est_level_std = 0
	est_level = np.zeros(len(test_dataset))
	est_a = np.zeros(len(test_dataset))
	est_b = np.zeros(len(test_dataset))

	with torch.no_grad():
		for i, (images, targets) in enumerate(tqdm(test_loader, leave=False)):
			(image_clean, image_deg) = images
			(labels, levels) = targets
			image_clean=image_clean.to(device)
			image_deg=image_deg.to(device)
			labels=labels.to(device)
			levels=levels.float()
			levels=levels.to(device)
			if d["feature_loss"] == "COS":
				dum = torch.ones((image_clean.size(0),))
				dum = dum.to(device)
			# Predictions
			if deg_flag is None:
				# Evaluate the clean model by using degraded images
				msg = 'Not supported for simple calssifiers.'
				raise TypeError(msg)
			elif deg_flag == 0:
				msg = 'Not supported for distillation approaches.'
				raise TypeError(msg)
			elif deg_flag == 1:
				if d["deg_lev_est"]:
					coef_a, coef_b, deg_lev = net(image_clean, image_deg, True)
				else:
					coef_a, coef_b = net(image_clean, image_deg, True)

			if d["deg_lev_est"]:
				est_level[i] = deg_lev.item()
			est_a[i] = coef_a.item()
			est_b[i] = coef_b.item()

		# Results for each degradation level
		val_est_a = est_a.mean()
		val_est_b = est_b.mean()
		val_est_a_std = np.std(est_a, ddof = 1)
		val_est_b_std = np.std(est_b, ddof = 1)
		if d["deg_lev_est"]:
			val_est_level = est_level.mean()
			val_est_level_std = np.std(est_level, ddof = 1)

	resultlist=[
				lev,
				val_est_a,
				val_est_b,
				val_est_level,
				val_est_a_std,
				val_est_b_std,
				val_est_level_std
	]
	save_csv(root_dir = dir, file_name = filename, target_list = resultlist, header_list = headerlist)
	print(*resultlist)

# Unfix the seeds of Guassian and binamial distributions for degradation
degradedimagedata.fix_seed_noise_sl(is_fixed = False)
print("----- Evaluation END -----")
