"""
Evaluate Degradation Removal Adapter
"""
import sys
import torch
import torch.nn as nn
import torchvision
from torchvision import transforms
from tqdm import tqdm
import json
import numpy as np
import copy
import torch.nn.functional as F

from modelforcifar import dra_nwbuilder
from datahandler import degradedimagedata
from dra_common import save_model, load_model, save_csv, get_logmodel_dir, num_correctness, timestamp, length_and_degree, get_eval_dir_filename

# Sys args
last_epoch = sys.argv[1] # Last epoch, should be "0" in the case of initial training.
exp_string = sys.argv[2] # Select a key string from "CLEAN", "MSE" and "COS"
exp_no = int(sys.argv[3]) # 0: CLEAN, 1-4: MSE, 1-8: COS

# Meta settings
batch_size = 64
num_workers = 4
simul = False # False if sequential or Ture if simultenous
headerlist=["Deg_level","CRL","CEL_clean","CEL_deg","DLL","ACC_clean","ACC_deg",
			"Est_level","Est_level_std"]

# Configuration for experiments
model_flag = 3
open_config = open('experiments/resnet50cub2011_config.json','r')

try:
	model_config = json.load(open_config)
finally:
	open_config.close()

d = model_config[exp_string][exp_no]
#deg_type, deg_range = degradedimagedata.get_type_range(d["deg_type"])
deg_type, deg_list = degradedimagedata.get_type_list(d["deg_type"])
deg_flag = d["deg_flag"]

dir, filename = get_eval_dir_filename(deg_type = d["deg_type"], model_name = d["model_name"], deg_flag = deg_flag, deg_lev_est = d["deg_lev_est"], feature_loss = d["feature_loss"], simul = simul, is_feat_only = d["is_feat_only"], epoch = last_epoch, exp_string = exp_string)

# load a model
device = 'cuda' if torch.cuda.is_available() else 'cpu'
root_dir = get_logmodel_dir(deg_type = d["deg_type"], model_name = d["model_name"], deg_flag = deg_flag, deg_lev_est = d["deg_lev_est"], feature_loss = d["feature_loss"], simul = simul, exp_string = exp_string)
last_checkpoint = load_model(root_dir = root_dir, epoch = last_epoch)
dra_nw = dra_nwbuilder.buildNetwork(model_name = d["model_name"], num_of_classes = d["num_classes"], num_of_features = d["num_features"], init_weights = True, deg_flag = d["deg_flag"], is_deglev_estimated = d["deg_lev_est"])
dra_nw.load_state_dict(last_checkpoint["model_state_dict"])
net = dra_nw.to(device)
net.eval()

# Loss function
cross_entoropy_loss_clean = nn.CrossEntropyLoss()
cross_entoropy_loss_deg = nn.CrossEntropyLoss()
if d["feature_loss"] == "MSE":
	consistency_loss = nn.MSELoss()
elif d["feature_loss"] == "COS":
	consistency_loss = nn.CosineEmbeddingLoss()
elif d["feature_loss"] == "KLD":
	consistency_loss = nn.KLDivLoss(reduction='batchmean')
elif d["feature_loss"] == "DIS":
	consistency_loss = nn.KLDivLoss(reduction='batchmean')
elif d["feature_loss"] is not None:
	msg = 'The type of consitency loss is not supported.'
	raise LookupError(msg)

if d["deg_lev_est"]:
	deg_levels_loss = nn.MSELoss()

# Fix the seeds of Guassian and binomial distributions for degradation
degradedimagedata.fix_seed_noise_sl(is_fixed = True)

# Transform image dataset
# For clean
test_transform = transforms.Compose([transforms.Resize(size = 256)])

print("----- Evaluation START -----")
print(*headerlist)

for lev in deg_list:
	# Data generation
	test_dataset = degradedimagedata.get_datasetclass(data_name = d["data_name"], train = False, transform = test_transform, deg_type = deg_type, deg_range = None, deg_list = [lev, lev], download = False, is_resized_before_deg = False, cgn_traindata_gen = True)
	test_loader = torch.utils.data.DataLoader( test_dataset, batch_size = batch_size, shuffle = False, num_workers = num_workers)

	# Initialize variabels
	val_loss = 0
	val_consistency_loss = 0
	val_cross_entoropy_loss_clean = 0
	val_cross_entoropy_loss_deg = 0
	val_deg_lev_loss = 0
	val_acc_clean = 0
	val_acc_deg = 0
	val_est_level = 0
	val_est_level_std = 0

	with torch.no_grad():
		for i, (images, targets) in enumerate(tqdm(test_loader, leave=False)):
			(image_clean, image_deg) = images
			(labels, levels) = targets
			image_clean=image_clean.to(device)
			image_deg=image_deg.to(device)
			labels=labels.to(device)
			levels=levels.float()
			levels=levels.to(device)
			val_num_of_samples = labels.size(0)
			if d["feature_loss"] == "COS":
				dum = torch.ones((image_clean.size(0),))
				dum = dum.to(device)
			# Predictions
			if deg_flag is None:
				# Evaluate the clean model by using degraded images
				clean_out, clean_feat = net(image_deg, image_deg)
			elif deg_flag == 0:
				if d["deg_lev_est"]:
					clean_out, clean_feat, deg_out, deg_feat, deg_lev = net(image_clean, image_deg)
				else:
					clean_out, clean_feat, deg_out, deg_feat = net(image_clean, image_deg)
				val_acc_deg += num_correctness(deg_out, labels)
			elif deg_flag == 1:
				if d["deg_lev_est"]:
					clean_out, clean_feat, deg_out, deg_feat, deg_lev = net(image_clean, image_deg)
				else:
					clean_out, clean_feat, deg_out, deg_feat = net(image_clean, image_deg)
				val_acc_deg += num_correctness(deg_out, labels)

			val_acc_clean += num_correctness(clean_out, labels)

			# Calculate losses
			if deg_flag is None:
				ce_clean = cross_entoropy_loss_clean(clean_out, labels)
				val_cross_entoropy_loss_clean += ce_clean.item()*val_num_of_samples
			else:
				if d["feature_loss"] == "COS":
					cl = consistency_loss(clean_feat, deg_feat, dum)
				elif d["feature_loss"] == "KLD":
					clean_task_metric = copy.deepcopy(net.clean_task)
					clean_task_metric.eval()
					for param_metric in clean_task_metric.parameters():
						param_metric.requires_grad = False
					pred_metric = clean_task_metric(deg_feat)
					cl = consistency_loss(torch.log(torch.clamp(F.softmax(pred_metric,dim=1),1e-10,1.0)), torch.clamp(F.softmax(clean_out,dim=1),1e-10,1.0))
				elif d["feature_loss"] == "DIS":
					cl = consistency_loss(torch.log(torch.clamp(F.softmax(deg_out,dim=1),1e-10,1.0)), torch.clamp(F.softmax(clean_out,dim=1),1e-10,1.0))
				else:
					cl = consistency_loss(clean_feat, deg_feat)
				val_consistency_loss += cl.item()*val_num_of_samples
				if d["deg_lev_est"]:
					dl = deg_levels_loss(deg_lev, levels)
					val_deg_lev_loss += dl.item()*val_num_of_samples
				ce_deg = cross_entoropy_loss_deg(deg_out, labels)
				val_cross_entoropy_loss_deg += ce_deg.item()*val_num_of_samples
				ce_clean = cross_entoropy_loss_clean(clean_out, labels)
				val_cross_entoropy_loss_clean += ce_clean.item()*val_num_of_samples

			if d["deg_lev_est"]:
				val_est_level += torch.sum(deg_lev).item()
		# Results for each degradation level
		val_consistency_loss /= len(test_loader.dataset)
		val_cross_entoropy_loss_clean /= len(test_loader.dataset)
		val_cross_entoropy_loss_deg /= len(test_loader.dataset)
		val_deg_lev_loss /= len(test_loader.dataset)
		val_acc_clean /= len(test_loader.dataset)
		val_acc_deg /= len(test_loader.dataset)
		val_est_level /= len(test_loader.dataset)

	resultlist=[
				lev,
				val_consistency_loss,
				val_cross_entoropy_loss_clean,
				val_cross_entoropy_loss_deg,
				val_deg_lev_loss,
				val_acc_clean,
				val_acc_deg,
				val_est_level,
				val_est_level_std
	]
	save_csv(root_dir = dir, file_name = filename, target_list = resultlist, header_list = headerlist)
	print(*resultlist)

# Unfix the seeds of Guassian and binamial distributions for degradation
degradedimagedata.fix_seed_noise_sl(is_fixed = False)
print("----- Evaluation END -----")
