"""
Train Degradation Removal Adapter
"""
import sys
import torch
import torch.nn as nn
import torchvision
from torchvision import transforms
from tqdm import tqdm
from torchsummary import summary # https://github.com/sksq96/pytorch-summary
import json
import copy
import csv
import torch.nn.functional as F

from modelforcifar import dra_nwbuilder
from datahandler import degradedimagedata
from dra_common import save_model, load_model, save_csv, get_logmodel_dir, num_correctness, timestamp
from radam import RAdam
from datahandler import scaleaugtransforms

# Sys args
num_epochs = int(sys.argv[1]) # Number of epochs
last_epoch = int(sys.argv[2]) # Last epoch, should be "0" in the case of initial training.
exp_string = sys.argv[3] # Select a key string from "CLEAN", "MSE" and "COS"
exp_no = int(sys.argv[4]) # 0: CLEAN, 0: DEG, 1-10: MSE, 1-10: COS, 1-10: KLD
view_model = sys.argv[5] # view only if "view"

# Meta settings
batch_size = 64# CIFAR:128, CUB:64
test_batch_size = 64# CIFAR:100, CUB:64
num_workers = 4
simul = False # False if sequential or Ture if simultenous
headerlist=["Epoch","Time","Train_loss","Train_CRL","Train_CEL_clean","Train_CEL_deg","Train_DLL","Train_ACC_clean","Train_ACC_deg",
			"Val_loss","Val_CRL","Val_CEL_clean","Val_CEL_deg","Val_DLL","Val_ACC_clean","Val_ACC_deg"]
max_val_acc = 0.1
min_val_loss = 1.0
optimizer_reset = False
is_lrSchedule = True
is_dataparallel = False
use_pretrained = True

# Configuration for experiments
model_flag = 3
open_config = open('experiments/resnet50cub2011_config.json','r')

try:
	model_config = json.load(open_config)
finally:
	open_config.close()

d = model_config[exp_string][exp_no]
weight_config = model_config["WEIGHTS"][exp_no]
ce_clean_w = weight_config["ce_clean_w"]
ce_deg_w = weight_config["ce_deg_w"]
cl_w = weight_config["cl_w"]
dl_w = weight_config["dl_w"]
deg_list = None
deg_range = None
deg_type, deg_list = degradedimagedata.get_type_list(d["deg_type"])

# Data generation
# For clean
transform = transforms.Compose([transforms.Resize(size = 256)])
train_dataset = degradedimagedata.get_datasetclass(data_name = d["data_name"], train = True, transform = transform, deg_type = deg_type, deg_range = deg_range, deg_list = deg_list, download = False, is_resized_before_deg = False, cgn_traindata_gen = True)
train_loader = torch.utils.data.DataLoader( train_dataset, batch_size = batch_size, shuffle = True, num_workers = num_workers)

# For clean
test_transform = transforms.Compose([transforms.Resize(size = 256)])
test_dataset = degradedimagedata.get_datasetclass(data_name = d["data_name"], train = False, transform = test_transform, deg_type = deg_type, deg_range = deg_range, deg_list = deg_list, download = False, is_resized_before_deg = False, cgn_traindata_gen = True)
test_loader = torch.utils.data.DataLoader( test_dataset, batch_size = test_batch_size, shuffle = False, num_workers = num_workers)

# Build or load a model
device = 'cuda' if torch.cuda.is_available() else 'cpu'
deg_flag = d["deg_flag"]

root_dir = get_logmodel_dir(deg_type = d["deg_type"], model_name = d["model_name"], deg_flag = deg_flag, deg_lev_est = d["deg_lev_est"], feature_loss = d["feature_loss"], simul = simul, exp_string = exp_string)
last_checkpoint = load_model(root_dir = root_dir, epoch = last_epoch)

# Model initilization
if last_checkpoint is not None:
	# Continue the training.
	dra_nw = dra_nwbuilder.buildNetwork(model_name = d["model_name"], num_of_classes = d["num_classes"], num_of_features = d["num_features"], init_weights = True, deg_flag = d["deg_flag"], is_deglev_estimated = d["deg_lev_est"])
	dra_nw.load_state_dict(last_checkpoint["model_state_dict"])
	# If you train the clean model, you have to activate the following two lines.
	#dra_nw.train_clean_features_partially_startswith(num_list = ['0.', '4.', '5.conv1'], is_training = False)
	#dra_nw.train_clean_features_partially_in(num_list = ['.bn'], is_training = True)

	if (not simul) and ((deg_flag == 0) or (deg_flag == 1)):
		dra_nw.train_clean_features(is_training = False)
		dra_nw.train_clean_task(is_training = False)
elif simul:
	# Simultenous training
	dra_nw = dra_nwbuilder.buildNetwork(model_name = d["model_name"], num_of_classes = d["num_classes"], num_of_features = d["num_features"], init_weights = True, deg_flag = d["deg_flag"], is_deglev_estimated = d["deg_lev_est"])
	dra_nw.init_adapter(init_weights = True, use_clean = False)
	dra_nw.init_deglev_estimated(init_weights = True)
elif ((deg_flag == 0) or (deg_flag == 1)):
	# Training will start with a pre-trained clean model.
	# Loading clean parameters
	dra_nw = dra_nwbuilder.buildNetwork(model_name = d["model_name"], num_of_classes = d["num_classes"], num_of_features = d["num_features"], init_weights = True, deg_flag = None, is_deglev_estimated = False)
	clean_root_dir = get_logmodel_dir(deg_type = d["deg_type"], model_name = d["model_name"], deg_flag = None, deg_lev_est = False)
	clean_checkpoint = load_model(root_dir = clean_root_dir, epoch = last_epoch)
	dra_nw.load_state_dict(clean_checkpoint['model_state_dict'])
	# Making degradation removal adapter
	dra_nw.init_adapter(init_weights = True, use_clean = True, deg_flag = deg_flag)
	if d["deg_lev_est"]:
		dra_nw.set_deglev_estimated(is_deglev_estimated = True)
		dra_nw.init_deglev_estimated(init_weights = True)
	# Freeze all the parameters of clean model
	dra_nw.train_clean_features(is_training = False)
	dra_nw.train_clean_task(is_training = False)
else:
	dra_nw = dra_nwbuilder.buildNetwork(model_name = d["model_name"], num_of_classes = d["num_classes"], num_of_features = d["num_features"], init_weights = True, deg_flag = None, is_deglev_estimated = False)
	#for param_tensor in dra_nw.state_dict():
	#	print(param_tensor, "\t", dra_nw.state_dict()[param_tensor].size())
	if use_pretrained:
		print("----- START loading pretrained ResNet50 -----")
		pre_model = torchvision.models.resnet50(pretrained = True)
		pre_state_dict = copy.deepcopy(pre_model.state_dict())
		torch_name = []
		with open("torch_param_resnet.csv", 'r') as f:
			reader = csv.reader(f)
			for row in reader:
				torch_name += row
		my_name = []
		with open("my_param_resnet.csv", 'r') as f:
			reader = csv.reader(f)
			for row in reader:
				my_name += row
		init_checkpoint = copy.deepcopy(dra_nw.state_dict())
		for i in range(len(torch_name)):
			init_checkpoint[my_name[i]] = pre_state_dict[torch_name[i]]
		print("----- END loading pretrained ResNet50 -----")
		dra_nw.load_state_dict(init_checkpoint)
		print("----- Freeze features of pretrained ResNet50 -----")
		dra_nw.train_clean_features_partially_startswith(num_list = ['0.', '4.', '5.conv1'], is_training = False)
		dra_nw.train_clean_features_partially_in(num_list = ['.bn'], is_training = True)

# In the case of Sequetianl training
if (not simul) and ((deg_flag == 0) or (deg_flag == 1)):
	dra_nw.train_adp_features(is_training = True)
	if d["is_feat_only"]:
		dra_nw.train_deg_task(is_training = False)
	else:
		dra_nw.train_deg_task(is_training = True)

# Send a model to a device
net = dra_nw.to(device)
if is_dataparallel:
	net = nn.DataParallel(net, device_ids=[0,1])

optimizer = torch.optim.Adam(net.parameters(), lr = 1e-4) # For the network trained with clean images

if is_lrSchedule:
	scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=0.95) # 0.95 For the network trained with clean images 1st

if ((last_checkpoint is not None) and (not(optimizer_reset))):
	optimizer.load_state_dict(last_checkpoint["optimizer_state_dict"])

# Loss function
cross_entoropy_loss_clean = nn.CrossEntropyLoss()
cross_entoropy_loss_deg = nn.CrossEntropyLoss()
if d["feature_loss"] == "MSE":
	consistency_loss = nn.MSELoss()
elif d["feature_loss"] == "COS":
	consistency_loss = nn.CosineEmbeddingLoss()
elif d["feature_loss"] == "KLD":
	consistency_loss = nn.KLDivLoss(reduction='batchmean')
elif d["feature_loss"] == "DIS":
	consistency_loss = nn.KLDivLoss(reduction='batchmean')
elif d["feature_loss"] is not None:
	msg = 'The type of consitency loss is not supported.'
	raise LookupError(msg)

if d["deg_lev_est"]:
	deg_levels_loss = nn.MSELoss()

if view_model == "view":
	print('-----Model Summary-----')
	input_size = train_dataset.getCHW()
	summary(net, [input_size,input_size])
	sys.exit()

print('-----Training start at {}-----'.format(timestamp()))
print(*headerlist)
for epoch in range(num_epochs):
	train_loss = 0
	train_consistency_loss = 0
	train_cross_entoropy_loss_clean = 0
	train_cross_entoropy_loss_deg = 0
	train_deg_lev_loss = 0
	train_acc_clean = 0
	train_acc_deg = 0
	val_loss = 0
	val_consistency_loss = 0
	val_cross_entoropy_loss_clean = 0
	val_cross_entoropy_loss_deg = 0
	val_deg_lev_loss = 0
	val_acc_clean = 0
	val_acc_deg = 0

	# ===== Start training =====
	net.train()
	if ((deg_flag is not None) and not(simul)):# Disable upadating batchnorm paramters and dropout
		net.freeze_clean_features(is_training = False)
		net.freeze_clean_task(is_training = False)

	for i, (images, targets) in enumerate(tqdm(train_loader, leave=False)):
		(image_clean, image_deg) = images
		(labels, levels) = targets
		image_clean = image_clean.to(device)
		image_deg = image_deg.to(device)
		levels = levels.float()
		labels = labels.to(device)
		levels = levels.to(device)
		num_of_samples = labels.size(0)
		if d["feature_loss"] == "COS":
			dum = torch.ones((image_clean.size(0),))
			dum = dum.to(device)

		# Forward induction
		optimizer.zero_grad()
		if deg_flag == None:
			if exp_string == "CLEAN":
				clean_out, clean_feat = net(image_clean, image_deg)
			elif exp_string == "DEG":
				clean_out, clean_feat = net(image_deg, image_deg)
		elif deg_flag == 0:
			if d["deg_lev_est"]:
				clean_out, clean_feat, deg_out, deg_feat, deg_lev = net(image_clean, image_deg)
			else:
				clean_out, clean_feat, deg_out, deg_feat = net(image_clean, image_deg)
			train_acc_deg += num_correctness(deg_out, labels)
		elif deg_flag == 1:
			if d["deg_lev_est"]:
				clean_out, clean_feat, deg_out, deg_feat, deg_lev = net(image_clean, image_deg)
			else:
				clean_out, clean_feat, deg_out, deg_feat = net(image_clean, image_deg)
			train_acc_deg += num_correctness(deg_out, labels)

		train_acc_clean += num_correctness(clean_out, labels)

		# Backward induction
		if deg_flag == None:
			loss = cross_entoropy_loss_clean(clean_out, labels)
			train_cross_entoropy_loss_clean += loss.item()*num_of_samples
		else:
			if d["feature_loss"] == "COS":
				cl = consistency_loss(clean_feat, deg_feat, dum)
			elif d["feature_loss"] == "KLD":
				clean_task_metric = copy.deepcopy(net.clean_task)
				clean_task_metric.eval()
				for param_metric in clean_task_metric.parameters():
					param_metric.requires_grad = False
				pred_metric = clean_task_metric(deg_feat)
				cl = consistency_loss(torch.log(torch.clamp(F.softmax(pred_metric,dim=1),1e-10,1.0)), torch.clamp(F.softmax(clean_out,dim=1),1e-10,1.0))
			elif d["feature_loss"] == "DIS":
				cl = consistency_loss(torch.log(torch.clamp(F.softmax(deg_out,dim=1),1e-10,1.0)), torch.clamp(F.softmax(clean_out,dim=1),1e-10,1.0))
			else:
				cl = consistency_loss(deg_feat, clean_feat)
			train_consistency_loss += cl.item()*num_of_samples
			loss = cl_w * cl
			if d["deg_lev_est"]:
				dl = deg_levels_loss(deg_lev, levels)
				train_deg_lev_loss += dl.item()*num_of_samples
				loss += dl_w * dl

			ce_deg = cross_entoropy_loss_deg(deg_out, labels)
			train_cross_entoropy_loss_deg += ce_deg.item()*num_of_samples
			if (not d["is_feat_only"]):
				loss += ce_deg_w * ce_deg
			ce_clean = cross_entoropy_loss_clean(clean_out, labels)
			train_cross_entoropy_loss_clean += ce_clean.item()*num_of_samples

		train_loss += loss.item()*num_of_samples
		loss.backward()
		optimizer.step()
	# Results for each epoch
	train_loss /= len(train_loader.dataset)
	train_consistency_loss /= len(train_loader.dataset)
	train_cross_entoropy_loss_clean /= len(train_loader.dataset)
	train_cross_entoropy_loss_deg /= len(train_loader.dataset)
	train_deg_lev_loss /= len(train_loader.dataset)
	train_acc_clean /= len(train_loader.dataset)
	train_acc_deg /= len(train_loader.dataset)

	# ===== Start testing =====
	net.eval()
	with torch.no_grad():
		for i, (images, targets) in enumerate(test_loader):
			(image_clean, image_deg) = images
			(labels, levels) = targets
			image_clean = image_clean.to(device)
			image_deg = image_deg.to(device)
			labels = labels.to(device)
			levels = levels.float()
			levels = levels.to(device)
			val_num_of_samples = labels.size(0)
			if d["feature_loss"] == "COS":
				dum = torch.ones((image_clean.size(0),))
				dum = dum.to(device)
			# Predictions
			if deg_flag == None:
				if exp_string == "CLEAN":
					clean_out, clean_feat = net(image_clean, image_deg)
				elif exp_string == "DEG":
					clean_out, clean_feat = net(image_deg, image_deg)
			elif deg_flag == 0:
				if d["deg_lev_est"]:
					clean_out, clean_feat, deg_out, deg_feat, deg_lev = net(image_clean, image_deg)
				else:
					clean_out, clean_feat, deg_out, deg_feat = net(image_clean, image_deg)
				val_acc_deg += num_correctness(deg_out, labels)
			elif deg_flag == 1:
				if d["deg_lev_est"]:
					clean_out, clean_feat, deg_out, deg_feat, deg_lev = net(image_clean, image_deg)
				else:
					clean_out, clean_feat, deg_out, deg_feat = net(image_clean, image_deg)
				val_acc_deg += num_correctness(deg_out, labels)

			val_acc_clean += num_correctness(clean_out, labels)

			# Calculate losses
			if deg_flag == None:
				ce_clean = cross_entoropy_loss_clean(clean_out, labels)
				val_cross_entoropy_loss_clean += ce_clean.item()*val_num_of_samples
				loss = ce_clean
			else:
				if d["feature_loss"] == "COS":
					cl = consistency_loss(clean_feat, deg_feat, dum)
				elif d["feature_loss"] == "KLD":
					clean_task_metric = copy.deepcopy(net.clean_task)
					clean_task_metric.eval()
					for param_metric in clean_task_metric.parameters():
						param_metric.requires_grad = False
					pred_metric = clean_task_metric(deg_feat)
					cl = consistency_loss(torch.log(torch.clamp(F.softmax(pred_metric,dim=1),1e-10,1.0)), torch.clamp(F.softmax(clean_out,dim=1),1e-10,1.0))
				elif d["feature_loss"] == "DIS":
					cl = consistency_loss(torch.log(torch.clamp(F.softmax(deg_out,dim=1),1e-10,1.0)), torch.clamp(F.softmax(clean_out,dim=1),1e-10,1.0))
				else:
					cl = consistency_loss(deg_feat, clean_feat)
				val_consistency_loss += cl.item()*val_num_of_samples
				loss = cl_w * cl
				if d["deg_lev_est"]:
					dl = deg_levels_loss(deg_lev, levels)
					val_deg_lev_loss += dl.item()*val_num_of_samples
					loss += dl_w * dl
				ce_deg = cross_entoropy_loss_deg(deg_out, labels)
				val_cross_entoropy_loss_deg += ce_deg.item()*val_num_of_samples
				if (not d["is_feat_only"]):
					loss += ce_deg_w * ce_deg
				ce_clean = cross_entoropy_loss_clean(clean_out, labels)
				val_cross_entoropy_loss_clean += ce_clean.item()*val_num_of_samples
			val_loss += loss.item()*val_num_of_samples
		# Results for each epoch
		val_loss /= len(test_loader.dataset)
		val_consistency_loss /= len(test_loader.dataset)
		val_cross_entoropy_loss_clean /= len(test_loader.dataset)
		val_cross_entoropy_loss_deg /= len(test_loader.dataset)
		val_deg_lev_loss /= len(test_loader.dataset)
		val_acc_clean /= len(test_loader.dataset)
		val_acc_deg /= len(test_loader.dataset)

	loglist=[
		epoch+1+last_epoch,
		timestamp(),
		train_loss,
		train_consistency_loss,
		train_cross_entoropy_loss_clean,
		train_cross_entoropy_loss_deg,
		train_deg_lev_loss,
		train_acc_clean,
		train_acc_deg,
		val_loss,
		val_consistency_loss,
		val_cross_entoropy_loss_clean,
		val_cross_entoropy_loss_deg,
		val_deg_lev_loss,
		val_acc_clean,
		val_acc_deg
		]
	save_csv(root_dir = root_dir, file_name = "train_log.csv", target_list = loglist, header_list = headerlist)
	print(*loglist)
	if exp_no == 0:
		target_val_acc = val_acc_clean
	else:
		target_val_acc = val_acc_deg
	if max_val_acc < target_val_acc:
		max_val_acc = target_val_acc
		if is_dataparallel:
			save_model(root_dir = root_dir, model_name = 'dramodel', epoch = 'maxacc', target_dict = {"model_state_dict":net.module.state_dict(),"optimizer_state_dict":optimizer.state_dict()})
		else:
			save_model(root_dir = root_dir, model_name = 'dramodel', epoch = 'maxacc', target_dict = {"model_state_dict":net.state_dict(),"optimizer_state_dict":optimizer.state_dict()})
	if val_loss < min_val_loss:
		min_val_loss = val_loss
		if is_dataparallel:
			save_model(root_dir = root_dir, model_name = 'dramodel', epoch = 'minloss', target_dict = {"model_state_dict":net.module.state_dict(),"optimizer_state_dict":optimizer.state_dict()})
		else:
			save_model(root_dir = root_dir, model_name = 'dramodel', epoch = 'minloss', target_dict = {"model_state_dict":net.state_dict(),"optimizer_state_dict":optimizer.state_dict()})
	if is_lrSchedule:
		scheduler.step()

if is_dataparallel:
	save_model(root_dir = root_dir, model_name = 'dramodel', epoch = num_epochs + last_epoch, target_dict = {"model_state_dict":net.module.state_dict(),"optimizer_state_dict":optimizer.state_dict()})
else:
	save_model(root_dir = root_dir, model_name = 'dramodel', epoch = num_epochs + last_epoch, target_dict = {"model_state_dict":net.state_dict(),"optimizer_state_dict":optimizer.state_dict()})
