"""
Degradation Removal Adapter (Feature Adjustor)
"""
import torch
import torch.nn as nn
import copy

class DraBase(nn.Module):
	"""
	Degradation Removal Adapter Base Class
	Two methods should be implemeted; _make_layers_for_features and _make_layers_for_task.
	"""
	def __init__(self, init_weights = True, num_of_features = 0, deg_flag = None, is_deglev_estimated = False):
		"""
		Constructor
		Args:
			num_of_features (int) : The number of features extracted by the feature extractor.
			init_weights (bool) : True if you initialize weights.
			deg_flag (int) : None clean only, 0 simple distillation, 1 degradation removal adapter.
			is_deglev_estimated (bool) : True in multi-task learning of degradation levels.
		"""
		super(DraBase, self).__init__()
		self.num_of_features = num_of_features
		self.clean_features = self._make_layers_for_features()
		self.clean_task = self._make_layers_for_task()
		self.deg_flag = deg_flag
		self.adp_features = None
		self.adp_a = None
		self.adp_b = None
		self.deg_task = None
		self.is_deglev_estimated = is_deglev_estimated
		self.deglev_estimator = None
		self.init_adapter(init_weights) # Build Adapter
		self.init_deglev_estimated(init_weights) # Bulid a restimation of degradation levels
		if init_weights:
			self.clean_features.apply(self._initilize_weights_for_layer)
			self.clean_task.apply(self._initilize_weights_for_layer)

	def init_adapter(self, init_weights, use_clean = False, deg_flag = None):
		"""
		Initialize class variables for the adapter
		Args:
			init_weights (bool) : True if you initialize weights.
			use_clean (bool) : True if you initialize parameters with clean images' parameters.
			deg_flag (int) : None clean only, 0 simple distillation, 1 degradation removal adapter.
		"""
		if deg_flag is not None:
			self.deg_flag = deg_flag

		if self.deg_flag == 0:
			self.adp_features = self._make_layers_for_features()
			self.deg_task = self._make_layers_for_task()
			if init_weights:
				if use_clean:
					self.load_clean_weights()
				else:
					self.adp_features.apply(self._initilize_weights_for_layer)
					self.deg_task.apply(self._initilize_weights_for_layer)
		elif self.deg_flag == 1:
			self.adp_features = self._make_layers_for_features()
			self.adp_a = self._make_layers_for_deg('adp_a')
			self.adp_b = self._make_layers_for_deg('adp_b')
			self.deg_task = self._make_layers_for_task()
			if init_weights:
				self.adp_a.apply(self._initilize_weights_for_layer)
				self.adp_b.apply(self._initilize_weights_for_layer)
				if use_clean:
					self.load_clean_weights()
				else:
					self.adp_features.apply(self._initilize_weights_for_layer)
					self.deg_task.apply(self._initilize_weights_for_layer)
		else:
			self.adp_features = None
			self.adp_a = None
			self.adp_b = None
			self.deg_task = None

	def init_deglev_estimated(self, init_weights):
		"""
		Initialize class variables for the degradation level estimator
		Args:
			init_weights (bool) : True if you initialize weights.
		"""
		if self.is_deglev_estimated:
			self.deglev_estimator =self._make_layers_for_deg('deg_lev')
			if init_weights:
				self.deglev_estimator.apply(self._initilize_weights_for_layer)

	def set_deglev_estimated(self, is_deglev_estimated = True):
		"""
		Enable the degradation level estimation
		Args:
		 	is_deglev_estimated (bool) : True if estimation
		"""
		self.is_deglev_estimated = is_deglev_estimated

	def load_clean_weights(self):
		"""
		Load weights of clean CNN into the degrdation removal adapter
		"""
		self.adp_features.load_state_dict(self.clean_features.state_dict())
		self.deg_task.load_state_dict(self.clean_task.state_dict())

	def train_clean_features(self, is_training):
		"""
		Make the features of clean CNN trainable or not
		Args:
			is_training (bool) : True if trainable
		"""
		for param in self.clean_features.parameters():
			param.requires_grad = is_training

	def freeze_clean_features(self, is_training):
		"""
		Freeze all layers of clean features including Bactchnorm layers and dropout
		Args:
			is_training (bool) : True if trainable
		"""
		if is_training:
			self.clean_features.train()
		else:
			self.clean_features.eval()

	def train_clean_task(self, is_training):
		"""
		Make the task of clean CNN trainable or not
		Args:
			is_training (bool) : True if trainable
		"""
		for param in self.clean_task.parameters():
			param.requires_grad = is_training

	def freeze_clean_task(self, is_training):
		"""
		Freeze all layers of clean task including Bactchnorm layers and dropout
		Args:
			is_training (bool) : True if trainable
		"""
		if is_training:
			self.clean_task.train()
		else:
			self.clean_task.eval()

	def train_adp_features(self, is_training):
		"""
		Make the features of adapter CNN trainable or not
		Args:
			is_training (bool) : True if trainable
		"""
		for param in self.adp_features.parameters():
			param.requires_grad = is_training

	def train_deg_task(self, is_training):
		"""
		Make the task of degraded CNN trainable or not
		Args:
			is_training (bool) : True if trainable
		"""
		for param in self.deg_task.parameters():
			param.requires_grad = is_training

	def get_deg_flag(self):
		"""
		Existence of degradation removal adapter
		"""
		return self.deg_flag

	def is_degradation_level(self):
		"""
		Existence of degradation level estimator
		"""
		return self.is_deglev_estimated

	def forward(self, x_clean, x_deg, is_ab = False):
		"""
		Forward
		Please note the outputs form which depends on the exsistence of layers.
		Args:
			x_clean (tensor) : Input clean images.
			x_deg (tensor) : Input degraded images.
			is_ab (bool) : True if coefficients a and b are output.
		"""
		clean_img = x_clean
		deg_img = x_deg
		clean_feat = self.clean_features(clean_img)
		clean_feat = torch.flatten(clean_feat, 1)
		clean_out = self.clean_task(clean_feat)
		if self.deg_flag == 0:
			adp_feat = self.adp_features(deg_img)
			deg_feat = torch.flatten(adp_feat, 1)
			deg_out = self.deg_task(deg_feat)
			if self.is_deglev_estimated:
				deg_lev = self.deglev_estimator(deg_feat)
				fwd_out = clean_out, clean_feat, deg_out, deg_feat, deg_lev
			else:
				fwd_out = clean_out, clean_feat, deg_out, deg_feat
		elif self.deg_flag == 1:
			adp_feat = self.adp_features(deg_img)
			adp_feat = torch.flatten(adp_feat, 1)
			deg_a = self.adp_a(adp_feat)
			deg_b = self.adp_b(adp_feat)
			clean_feat_detach = copy.deepcopy(self.clean_features)
			clean_feat_detach = clean_feat_detach(deg_img).clone().detach()
			clean_feat_detach = torch.flatten(clean_feat_detach, 1)
			deg_feat = torch.mul(clean_feat_detach, deg_a)
			deg_feat = torch.add(deg_feat, deg_b)
			deg_out = self.deg_task(deg_feat)
			if self.is_deglev_estimated:
				deg_lev = self.deglev_estimator(adp_feat)
				if is_ab:
					fwd_out = torch.mean(deg_a, 1, True), torch.mean(deg_b, 1, True), deg_lev
				else:
					fwd_out = clean_out, clean_feat, deg_out, deg_feat, deg_lev
			else:
				if is_ab:
					fwd_out = torch.mean(deg_a, 1, True), torch.mean(deg_b, 1, True)
				else:
					fwd_out = clean_out, clean_feat, deg_out, deg_feat
		else:
			fwd_out = clean_out, clean_feat
		return fwd_out

	def _make_layers_for_features(self):
		"""
		Make feature extractor layers
		"""
		msg = "Feature layers have not been implemeted."
		raise NotImplementedError(msg)
		return None

	def _make_layers_for_task(self):
		"""
		Make specific task layers
		"""
		msg = "Feature layers have not been implemeted."
		raise NotImplementedError(msg)
		return None

	def _make_layers_for_deg(self, layer_name):
		"""
		Make layers for degradation removal adapter and degradation level estimator
		Args:
			layer_name (string) : layer name (adp_a, adp_b and deg_lev)
		"""
		layers = []
		if layer_name == 'adp_a':
			layers += [nn.Linear(self.num_of_features, self.num_of_features), nn.ReLU(inplace=True), nn.Linear(self.num_of_features, self.num_of_features), nn.Sigmoid()]
		elif layer_name == 'adp_b':
			layers += [nn.Linear(self.num_of_features, self.num_of_features), nn.ReLU(inplace=True), nn.Linear(self.num_of_features, self.num_of_features), nn.ReLU(inplace=True)]
		elif layer_name == 'deg_lev':
			layers += [nn.Linear(self.num_of_features, self.num_of_features), nn.ReLU(inplace=True), nn.Linear(self.num_of_features, self.num_of_features), nn.ReLU(inplace=True), nn.Linear(self.num_of_features, 1)]
		else:
			msg = 'Unsupported layer_name.'
			raise TypeError(msg)
		return nn.Sequential(*layers)

	def _initilize_weights_for_layer(self, layer):
		"""
		Initialize each layer depends on the layer type
		"""
		if isinstance(layer, nn.Conv2d):
			nn.init.kaiming_normal_(layer.weight, mode='fan_out', nonlinearity='relu')
			if layer.bias is not None:
				nn.init.constant_(layer.bias, 0)
		elif isinstance(layer, nn.BatchNorm2d):
			nn.init.constant_(layer.weight, 1)
			nn.init.constant_(layer.bias, 0)
		elif isinstance(layer, nn.Linear):
			nn.init.normal_(layer.weight, 0, 0.01)
			nn.init.constant_(layer.bias, 0)

	def train_clean_features_partially_startswith(self, num_list, is_training):
		"""
		Make the features of clean CNN trainable or not
		Args:
			num_list (list) : The list of layer numbers in clean_features
			is_training (bool) : True if trainable
		"""
		for target_num in num_list:
			target_string = str(target_num)
			for myname, param in self.clean_features.named_parameters():
				if myname.startswith(target_string):
					param.requires_grad = is_training

	def train_clean_features_partially_in(self, num_list, is_training):
		"""
		Make the features of clean CNN trainable or not
		Args:
			num_list (list) : The list of layer numbers in clean_features
			is_training (bool) : True if trainable
		"""
		for target_num in num_list:
			target_string = str(target_num)
			for myname, param in self.clean_features.named_parameters():
				if target_string in myname:
					param.requires_grad = is_training
