"""
Degradation Removal Adapter of ResNet
This code is an extenstion of Resnet codes seen in torchvision.models.resnet.
https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
"""
import torch
import torch.nn as nn
from modelforcifar import dra_base

cfgs = {
	'OrgResNet18': ['Basic', 2, 2, 2, 2],
	'OrgResNet34': ['Basic', 3, 4, 6, 3],
	'OrgResNet50': ['Bottleneck', 3, 4, 6, 3],
	'OrgResNet101': ['Bottleneck', 3, 4, 23, 3],
	'OrgResNet152': ['Bottleneck', 3, 8, 36, 3],
}

class BasicBlock(nn.Module):
	"""
	Basic block class
	"""
	expansion = 1
	def __init__(self, inplanes, planes, stride = 1, downsample = None, dilation = 1):
		"""
		Constructor
		Args:
			inplanes (int) : Input channels for the basic block
			planes (int) : Output channels for the basic block
			stride (int) : stride
			downsample (nn.Module) : downsample model
			dilation (int) : dilation rate
		"""
		super(BasicBlock, self).__init__()
		if dilation > 1:
			raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
		self.conv1 = nn.Conv2d(in_planes, planes, kernel_size = 3, stride = stride, padding = 1, bias = False)
		self.bn1 = nn.BatchNorm2d(planes)
		self.relu = nn.ReLU(inplace = True)
		self.conv2 = nn.Conv2d(planes, planes, kernel_size = 3, stride = 1, padding = 1, bias = False)
		self.bn2 = nn.BatchNorm2d(planes)
		self.downsample = downsample
		self.stride = stride

	def forward(self, x):
		identity = x
		out = self.relu(self.bn1(self.conv1(x)))
		out = self.bn2(self.conv2(out))
		if self.downsample is not None:
			identity = self.downsample(x)
		out += identity
		out = self.relu(out)
		return out

class Bottleneck(nn.Module):
	"""
	Bottleneck block class
	"""
	expansion = 4
	def __init__(self, inplanes, planes, stride = 1, downsample = None, dilation = 1):
		"""
		Constructor
		Args:
			inplanes (int) : Input channels for the bottleneck block
			planes (int) : Output channels for the bottleneck block
			stride (int) : stride
			downsample (nn.Module) : downsample model
			dilation (int) : dilation rate
		"""
		super(Bottleneck, self).__init__()
		self.conv1 = nn.Conv2d(inplanes, planes, kernel_size = 1, bias = False)
		self.bn1 = nn.BatchNorm2d(planes)
		self.conv2 = nn.Conv2d(planes, planes, kernel_size = 3, stride = stride, padding = 1, bias = False,  dilation = dilation)
		self.bn2 = nn.BatchNorm2d(planes)
		self.conv3 = nn.Conv2d(planes, self.expansion*planes, kernel_size = 1, bias = False)
		self.bn3 = nn.BatchNorm2d(self.expansion*planes)
		self.relu = nn.ReLU(inplace=True)
		self.stride = stride
		self.downsample = downsample

	def forward(self, x):
		identity = x
		out = self.relu(self.bn1(self.conv1(x)))
		out = self.relu(self.bn2(self.conv2(out)))
		out = self.bn3(self.conv3(out))
		if self.downsample is not None:
			identity = self.downsample(x)
		out += identity
		out = self.relu(out)
		return out

class DraResNet(dra_base.DraBase):
	"""
	Degradation Removal Adapter class of ResNet
	32 x 32 sized images are expected.
	"""
	def __init__(self, resnet_name, num_of_classes = 10, num_of_features = 512, init_weights = True, deg_flag = None, is_deglev_estimated = False):
		"""
		Constructor
		Args:
			resnet_name (string) : ResNet Type
			num_of_classes (int) : 10 for CIFAR-10, 100 for CIFAR-100.
			num_of_features (int) : The number of features extracted by the feature extractor.
			init_weights (bool) : True if you initialize weights.
			deg_flag (int) : None clean only, 0 simple distillation, 1 degradation removal adapter.
			is_deglev_estimated (bool) : True in multi-task learning of degradation levels.
		"""
		self.resnet_name = resnet_name
		self.num_of_classes = num_of_classes
		self.inplanes = 64
		self.dilation = 1
		super(DraResNet, self).__init__(init_weights, num_of_features, deg_flag, is_deglev_estimated)

	def _make_layer(self, block, planes, blocks, stride = 1, dilate = False):
		"""
		Make layers by using the BasicBlock class or the Bottleneck class
		Args:
			block (type) : BasicBlock or Bottleneck class
			planes (int) : The number of output features
			blocks (int) : The number of the unit block
			stride (int) : Stride
			dilate (bool): dialtaion or not
		"""
		downsample = None
		previous_dilation = self.dilation
		if dilate:
			self.dilation *= stride
			stride = 1
		if stride != 1 or self.inplanes != planes * block.expansion:
			downsample = nn.Sequential(nn.Conv2d(self.inplanes, block.expansion*planes, kernel_size = 1, stride = stride, bias = False), nn.BatchNorm2d(block.expansion*planes))

		layers = [block(self.inplanes, planes, stride, downsample, previous_dilation)]
		self.inplanes = planes * block.expansion
		for _ in range(1, blocks):
			layers += [block(self.inplanes, planes, dilation = self.dilation)]

		return layers

	def _get_block(self):
		"""
		Get the type of block
		"""
		if cfgs[self.resnet_name][0] == 'Basic':
			return BasicBlock
		else:
			return Bottleneck

	def _make_layers_for_features(self):
		"""
		Make feature extractor layers
		"""
		cfg = cfgs[self.resnet_name]
		layers = [nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False), nn.BatchNorm2d(self.inplanes), nn.ReLU(inplace=True), nn.MaxPool2d(kernel_size=3, stride=2, padding=1)]

		planes = 64
		layers += self._make_layer(self._get_block(), planes = planes, blocks = cfg[1], stride = 1)
		planes *= 2
		for i in [2,3,4]:
			layers += self._make_layer(self._get_block(), planes = planes, blocks = cfg[i], stride = 2)
			planes *= 2

		layers += [nn.AdaptiveAvgPool2d((1, 1))]
		self.inplanes = 64

		return nn.Sequential(*layers)

	def _make_layers_for_task(self):
		"""
		Make specific task layers
		"""
		return nn.Sequential(nn.Linear(512 * self._get_block().expansion, self.num_of_classes))
