"""
Degradation Removal Adapter of VGG for CIFAR-10/100
This code is an extenstion of VGG codes seen in torchvision.models.vgg.
"""
import torch
import torch.nn as nn
from modelforcifar import dra_base

cfgs = {
	'VGG11': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
	'VGG13': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
	'VGG16': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
	'VGG19': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}

class DraVggCifar(dra_base.DraBase):
	"""
	Degradation Removal Adapter class of VGG for CIFAR-10/100
	32 x 32 sized images are expected.
	"""
	def __init__(self, vgg_name, num_of_classes = 10, num_of_features = 512, init_weights = True, deg_flag = None, is_deglev_estimated = False):
		"""
		Constructor
		Args:
			vgg_name (string) : VGG Type
			num_of_classes (int) : 10 for CIFAR-10, 100 for CIFAR-100.
			num_of_features (int) : The number of features extracted by the feature extractor.
			init_weights (bool) : True if you initialize weights.
			deg_flag (int) : None clean only, 0 simple distillation, 1 degradation removal adapter.
			is_deglev_estimated (bool) : True in multi-task learning of degradation levels.
		"""
		self.vgg_name = vgg_name
		self.num_of_classes = num_of_classes
		super(DraVggCifar, self).__init__(init_weights, num_of_features, deg_flag, is_deglev_estimated)

	def _make_layers_for_features(self):
		"""
		Make feature extractor layers
		"""
		layers = []
		in_channels = 3
		cfg = cfgs[self.vgg_name]
		for x in cfg:
			if x == 'M':
				layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
			else:
				layers += [nn.Conv2d(in_channels, x, kernel_size=3, padding=1),
							nn.BatchNorm2d(x),
							nn.ReLU(inplace=True)]
				in_channels = x
		return nn.Sequential(*layers)

	def _make_layers_for_task(self):
		"""
		Make specific task layers
		"""
		return nn.Sequential(nn.Linear(self.num_of_features, self.num_of_classes))
